
# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.2.100'
version = '1.2.100'
full_version = '1.2.100'
git_revision = 'b47924ade1b8299774a30874fa2b951fc8c244b6'
release = True

if not release:
    version = full_version
    short_version += ".dev"
